/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: VirtualFile.java,v 1.9 2005/08/31 12:28:45 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Date;
import java.util.Iterator;
import java.util.NoSuchElementException;

import org.fossilec.odettej.service.RecordFormat;

/**
 * 
 * @author Rafael Marins
 */
public class VirtualFile {

	/**
	 * This enumeration serves to identify Virtual File mapping mode, whether
	 * it's for inputing or outputing.
	 */
	public static final class MappingMode {

		public static final MappingMode INPUT_MAPPING = new MappingMode();

		public static final MappingMode OUTPUT_MAPPING = new MappingMode();

		private MappingMode() {
			super();
		}
	}

	/**
	 * 
	 */
	private static class RecordIterator implements Iterator {

		private VirtualFile virtualFile;

		private int recordNumber;

		private int recordCount;

		private RecordIterator(VirtualFile vf) {

			this.virtualFile = vf;
			this.recordNumber = vf.getOffset();

			try {
				this.recordCount = vf.getRecordCount();
			} catch (OdetteFTPException e) {
				e.printStackTrace();
			}
		}

		public boolean hasNext() {
			return (recordNumber < recordCount);
		}

		public Object next() {
			MappedByteBuffer record = null;

			if (hasNext()) {
				try {
					record = virtualFile.readData(recordNumber);
					recordNumber++;
				} catch (OdetteFTPException e) {
					e.printStackTrace();
				}
			} else {
				throw new NoSuchElementException();
			}

			return record;
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}

	}

	/**
	 * File size's block length in octets.
	 */
	private static final int BLOCK_SIZE = 1024;

	/**
	 * Dataset name of the Virtual File being transfered, assigned by bilateral
	 * agreement.
	 */
	private String fileDatasetName;

	/**
	 * A file qualifier indicating the date and time the Virtual File was made
	 * available for transmission.
	 */
	private Date fileDateTime;

	/**
	 * Odette FTP can negotiate the restart of an interrupted Virtual File
	 * transmission. Fixed and Variable format files are restarted on record
	 * boundaries. For Unstructured and Text files the restart position is
	 * expressed as a file offset in 1K (1024 octet) blocks. The restart
	 * position is always calculated relative to the Virtual File.
	 */
	private int fileOffset;

	private VirtualFileMappingStrategy mappingStrategy;

	private MappingMode fileMode;

	private RecordFormat fileRecordFormat;

	private int fileRecordSize;

	private int fileRecordCount = -1;

	private int fileSize = -1;

	private File file;

	private FileChannel channel;

	public VirtualFile(String path, MappingMode mode,
			RecordFormat recordFormat, int recordSize)
			throws OdetteFTPException {

		this(path, mode, recordFormat, recordSize, 0);
	}

	public VirtualFile(String path, MappingMode mode,
			RecordFormat recordFormat, int recordSize, int restartOffset)
			throws OdetteFTPException {

		super();

		file = loadFile(path);
		this.fileDatasetName = file.getName();
		this.fileMode = mode;

		setupRecordFormat(recordFormat, recordSize);
		openFile(restartOffset);
	}

	public VirtualFile(String path, MappingMode mode,
			RecordFormat recordFormat, int recordSize, String datasetName)
			throws OdetteFTPException {

		this(path, mode, recordFormat, recordSize, datasetName, 0);
	}

	public VirtualFile(String path, MappingMode mode,
			RecordFormat recordFormat, int recordSize, String datasetName,
			int restartOffset) throws OdetteFTPException {

		super();

		file = loadFile(path);
		this.fileDatasetName = datasetName;
		this.fileMode = mode;

		setupRecordFormat(recordFormat, recordSize);
		openFile(restartOffset);
	}

	/**
	 * Return the Dataset Name of the Virtual File being transfered, assigned by
	 * bilateral agreement.
	 * 
	 * @return <code>String</code> containing the Dataset Name.
	 */
	public String getDatasetName() {
		return fileDatasetName;
	}

	/**
	 * Return the file qualifier indicating the date and time the Virtual File
	 * was made avaiable for transmission.
	 * 
	 * @return <code>Date</code> representing the specific date and time which
	 *         the Virtual File was made avaiable.
	 */
	public Date getDateTime() {
		if (fileDateTime == null) {
			long time = file.lastModified();
			fileDateTime = new Date(time);
		}

		return fileDateTime;
	}

	public void setDateTime(Date date) throws OdetteFTPException {
		fileDateTime = date;
		file.setLastModified(date.getTime());
	}

	public RecordFormat getFormat() {
		return fileRecordFormat;
	}

	FileChannel getChannel() {
		return channel;
	}

	/**
	 * Return the offset representing the restart position of the Virtual File
	 * being transfered. The count represent the record number if the record
	 * format is <code>FIXED</code> or <code>VARIABLE</code>, or the offset
	 * in 1K (1024 octets) blocks in case being <code>UNSTRUCTURED</code> or
	 * <code>TEXTFILE</code>.
	 * 
	 * @return <code>int</code> offset position. <code>zero</code> if hasn't
	 *         been set.
	 */
	public int getOffset() {
		return fileOffset;
	}

	public MappingMode getMode() {
		return fileMode;
	}

	private VirtualFileMappingStrategy getMappingStrategy() {
		if (mappingStrategy == null) {
			mappingStrategy = factoryMappingStrategy(getFormat());
		}
		return mappingStrategy;
	}

	private VirtualFileMappingStrategy factoryMappingStrategy(
			RecordFormat recordFormat) {

		VirtualFileMappingStrategy strategy = null;

		strategy = new FixedFileMapping();

		strategy.setVirtualFile(this);

		return strategy;
	}

	/**
	 * Return the file size in 1K (1024 octets) blocks required at the
	 * Originator location to store the Virtual File.<br>
	 * 
	 * This parameter is intended to provide only a good estimate of the Virtual
	 * File size.
	 * 
	 * @return <code>int</code> File Size in 1K (1024 octets) blocks.
	 * @throws OdetteFTPException
	 */
	public int getFileSize() throws OdetteFTPException {
		if (fileSize == -1) {
			long length = getUnitCount();
			setupSizes(length);
		}

		return fileSize;
	}

	public int getRecordSize() {
		return fileRecordSize;
	}

	public long getUnitCount() throws OdetteFTPException {
		try {
			return channel.size();
		} catch (IOException e) {
			throw new VirtualFileMappingException("Error occur in "
					+ "calculating Virtual File size", e);
		}
	}

	public int getRecordCount() throws OdetteFTPException {

		if (fileRecordCount == -1) {
			long length = getUnitCount();
			setupSizes(length);
		}

		return fileRecordCount;
	}

	/**
	 * Convenient method for setting the file size specifying the total amount
	 * of octets which is converted into blocks.
	 * 
	 * @param octets
	 *            Count in octets of Virtual File length.
	 */
	private void setupSizes(long octets) {
		int records;
		int blocks;

		blocks = (int) (octets / BLOCK_SIZE);
		if ((octets % BLOCK_SIZE) > 0) {
			blocks++;
		}

		records = (int) (octets / getRecordSize());
		if ((octets % getRecordSize()) > 0) {
			records++;
		}

		this.fileSize = blocks;
		this.fileRecordCount = records;
	}

	private void resetSizes() {
		this.fileSize = -1;
		this.fileRecordCount = -1;
	}

	private void setupRecordFormat(RecordFormat recordFormat, int recordSize) {

		if ((recordFormat == RecordFormat.TEXTFILE)
				|| (recordFormat == RecordFormat.UNSTRUCTURED)) {
			recordSize = BLOCK_SIZE;
		}

		this.fileRecordFormat = recordFormat;
		this.fileRecordSize = recordSize;
	}

	private FileChannel openFile(int restartOffset) throws OdetteFTPException {

		if (fileMode == MappingMode.INPUT_MAPPING) {
			channel = openInputFileChannel(file);

		} else if (fileMode == MappingMode.OUTPUT_MAPPING) {
			channel = openOutputFileChannel(file);

		} else {
			throw new IllegalArgumentException("invalid mapping mode: "
					+ fileMode);
		}

		setupRestart(restartOffset);
		resetSizes();

		return channel;
	}

	private static FileChannel openInputFileChannel(File file)
			throws OdetteFTPException {

		FileChannel channel = null;
		FileInputStream inputFile;

		try {
			inputFile = new FileInputStream(file);
			channel = inputFile.getChannel();

		} catch (FileNotFoundException e) {
			throw new VirtualFileMappingException("File not found: "
					+ file.getPath(), e);
		}

		return channel;
	}

	private static FileChannel openOutputFileChannel(File file)
			throws OdetteFTPException {

		FileChannel channel = null;
		FileOutputStream outputFile;

		try {
			outputFile = new FileOutputStream(file, true);
			channel = outputFile.getChannel();

		} catch (FileNotFoundException e) {
			throw new VirtualFileMappingException("File not found: "
					+ file.getPath(), e);
		}

		return channel;
	}

	public void setupRestart(int restartOffset) throws OdetteFTPException {

		if (restartOffset >= 0) {

			/* file size must be greater than the offset block */
			if (restartOffset > getRecordCount()) {
				throw new IllegalArgumentException("Restart offset greater "
						+ "than expected: " + restartOffset);
			}

			/* truncate output file to the restart block position */
			if (fileMode == MappingMode.OUTPUT_MAPPING) {
				truncateChannel(restartOffset);
			}

			// positionChannel(restartOffset);

			fileOffset = restartOffset;
		} else {
			throw new IllegalArgumentException("Illegal restart offset: "
					+ restartOffset);
		}

	}

	// private void positionChannel(int record) throws OdetteFTPException {
	// long position = calculateRecordPosition(record, getRecordSize());
	//
	// try {
	// channel.position(position);
	// } catch (IOException e) {
	// throw new VirtualFileMappingException("Error setting restart "
	// + "position");
	// }
	// }

	private void truncateChannel(int record) throws OdetteFTPException {
		long position = calculateRecordPosition(record, getRecordSize());

		try {
			channel.truncate(position);
		} catch (IOException e) {
			throw new VirtualFileMappingException("Failed while truncating "
					+ "file at the restart position", e);
		}
	}

	private static long calculateRecordPosition(int offset, int recordSize) {
		long position = offset * recordSize;
		return position;
	}

	private File loadFile(String path) {
		File reference = new File(path);
		return reference;
	}

	/**
	 * positioned from the restarting position. only valid in INPUT mode.
	 * 
	 * @return
	 */
	Iterator getRecordIterator() {
		Iterator it = new RecordIterator(this);
		return it;
	}

	MappedByteBuffer readData(int offset) throws OdetteFTPException {

		long position;
		int recordLength;

		VirtualFileMappingStrategy mapping = getMappingStrategy();
		position = calculateRecordPosition(offset, getRecordSize());

		if (offset < (getRecordCount() - 1)) {
			recordLength = getRecordSize();
		} else {
			recordLength = (int) (getUnitCount() - position);
		}

		return mapping.readData(position, recordLength);
	}

	void writeData(ByteBuffer buffer) throws OdetteFTPException {
		VirtualFileMappingStrategy mapping;

		if (fileMode != MappingMode.OUTPUT_MAPPING) {
			throw new IllegalStateException("Write require the OUTPUT state.");
		}

		long position;

		try {
			position = channel.size();
		} catch (IOException e) {
			throw new VirtualFileMappingException(e);
		}

		mapping = getMappingStrategy();
		mapping.writeData(position, buffer);

		resetSizes();
		this.fileOffset = getRecordCount();
	}

	public boolean isOpen() {
		return (channel == null ? false : channel.isOpen());
	}

	public void close() throws OdetteFTPException {
		try {
			channel.close();
		} catch (IOException e) {
			throw new VirtualFileMappingException(e);
		}
	}

	protected void finalize() throws Throwable {
		if (isOpen()) {
			close();
		}

		super.finalize();
	}
}
