/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: BufferMappingStrategy.java,v 1.3 2005/08/15 16:43:37 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.service;

import java.nio.ByteBuffer;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.fossilec.odettej.OdetteFTPException;

/**
 * 
 * @author Rafael Marins
 */
abstract class BufferMappingStrategy {

	protected static final int MAX_SUBRECORD_LENGTH = 63;

	protected static final int MIN_SUBRECORD_LENGTH = 2;

	private DataExchangeBuffer deb;

	public BufferMappingStrategy(DataExchangeBuffer deb) {
		super();
		this.deb = deb;
	}

	public ByteBuffer getBuffer() throws OdetteFTPException {
		return deb.buffer;
	}

	public abstract int fill(ByteBuffer record, RecordFormat recordFormat,
			boolean lastRecord) throws OdetteFTPException;

	public ByteBuffer scrape(boolean compression, RecordFormat recordFormat)
			throws OdetteFTPException {

		int totalSize = 0;

		List flattenSubrecords = unwrapSubrecords(compression, recordFormat);

		Iterator it = flattenSubrecords.iterator();
		while (it.hasNext()) {
			ByteBuffer bb = (ByteBuffer) it.next();
			totalSize += bb.capacity();
		}

		ByteBuffer scraptedBuffer = joinFlattenBuffer(flattenSubrecords,
				totalSize);

		return scraptedBuffer;
	}

	private List unwrapSubrecords(boolean compression, RecordFormat recordFormat)
			throws OdetteFTPException {

		ByteBuffer buffer;
		ByteBuffer subrecord;
		List flattenSubrecords;

		boolean endOfRecord;
		boolean compressed;
		byte size;

		boolean addLineSeparator;

		buffer = getBuffer();
		flattenSubrecords = new LinkedList();

		String lineSeparator = System.getProperty("line.separator");
		ByteBuffer separatorBuffer = ByteBuffer.wrap(lineSeparator.getBytes());

		while (buffer.hasRemaining()) {

			/* extract values from the subrecord header octet */
			short subrecordHeader = BufferUtil.getUnsignedByte(buffer);

			endOfRecord = ((subrecordHeader & 0x80) == 0x80); // is first bit
			// set?
			compressed = ((subrecordHeader & 0x40) == 0x40); // is second bit
			// set?
			size = (byte) (subrecordHeader & 0x3f); // count lasting 6 bits

			addLineSeparator = (recordFormat == RecordFormat.TEXTFILE)
					&& (endOfRecord);

			/* create a list of subrecords of this data exchange buffer */
			if (size > 0) {
				/*
				 * Text Files can map a blank line with a zero-length subrecord
				 * indicating End Of Record flag. Also, the specs determines
				 * that a zero-length subrecord may appear anywhere within a
				 * data exchange buffer.
				 */

				/* slice subrecord from the buffer */
				int entryLimit = buffer.limit();
				int endPosition = buffer.position() + (compressed ? 1 : size);

				buffer.limit(endPosition);
				subrecord = buffer.slice();

				/* set new buffer position and back to the original limit */
				buffer.position(endPosition);
				buffer.limit(entryLimit);

				/* add subrecord to the list */
				int repetition = (compressed ? size : 1);
				addSubrecordToList(flattenSubrecords, subrecord, repetition);
			}

			/* add line separator buffer to the list */
			if (addLineSeparator) {
				addSubrecordToList(flattenSubrecords, separatorBuffer);
			}

		}

		return flattenSubrecords;
	}

	private static ByteBuffer joinFlattenBuffer(List list, int totalSize) {

		ByteBuffer completeBuffer = ByteBuffer.allocateDirect(totalSize);

		Iterator it = list.iterator();
		while (it.hasNext()) {
			ByteBuffer subrecord = (ByteBuffer) it.next();
			subrecord.rewind();
			completeBuffer.put(subrecord);
		}

		completeBuffer.rewind();
		return completeBuffer;
	}

	private static void addSubrecordToList(List list, ByteBuffer subrecord) {
		addSubrecordToList(list, subrecord, 1);
	}

	private static void addSubrecordToList(List list, ByteBuffer subrecord,
			int repeat) {

		for (int i = 0; i < repeat; i++) {
			list.add(subrecord);
		}
	}

	protected static byte createSubrecordHeader(boolean endOfRecord,
			boolean compression, int count) {

		int octet = 0;

		octet += (endOfRecord ? 0x80 : 0x00);
		octet += (compression ? 0x40 : 0x00);
		octet += (count & 0x3f);

		return (byte) octet;
	}

}
