/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: CommandFormat.java,v 1.7 2005/08/15 16:43:37 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.service;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.WeakHashMap;

/**
 * This is a package visible class that provides support to represent a command
 * format. Command is a sequence of plain fields, which type should be Numeric
 * or Alphanumeric, and format Fixed or Variable.
 */
class CommandFormat implements Serializable {

	/**
	 * A field containing fixed values. All allowable values for the field are
	 * enumerated in the command definition.
	 */
	public static final char FIXED_FORMAT = 'F';

	/**
	 * A field with variable values within a defined range. For example the
	 * <i>SFIDFSIZ</i> field may contain any integer value between 0000000 and
	 * 9999999.
	 */
	public static final char VARIABLE_FORMAT = 'V';

	/**
	 * An alphanumeric field of length <i>n</i> octets.
	 */
	public static final char ALPHANUMERIC_TYPE = 'X';

	/**
	 * A numeric field of length <i>n</i> octets. This type of attribute
	 * contains characters from the following set:
	 * 
	 * <pre>
	 *    The numerals:               0 to 9
	 *    The upper case letters:     A to Z
	 *    The following special set:  / - . &amp; ( ) space.
	 * </pre>
	 */
	public static final char NUMERIC_TYPE = '9';

	/**
	 * Start Session Ready Message command format.
	 */
	private static final CommandFormat SSRM_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "SSRMCMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "SSRMMSG", FIXED_FORMAT, ALPHANUMERIC_TYPE, 17),
					new Field(18, "SSRMCR", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1) });

	/**
	 * Start Session command format.
	 */
	private static final CommandFormat SSID_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "SSIDCMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "SSIDLEV", FIXED_FORMAT, NUMERIC_TYPE, 1),
					new Field(2, "SSIDCODE", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 25),
					new Field(27, "SSIDPSWD", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 8),
					new Field(35, "SSIDSDEB", VARIABLE_FORMAT, NUMERIC_TYPE, 5),
					new Field(40, "SSIDSR", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(41, "SSIDCMPR", FIXED_FORMAT, ALPHANUMERIC_TYPE,
							1),
					new Field(42, "SSIDREST", FIXED_FORMAT, ALPHANUMERIC_TYPE,
							1),
					new Field(43, "SSIDSPEC", FIXED_FORMAT, ALPHANUMERIC_TYPE,
							1),
					new Field(44, "SSIDCRED", VARIABLE_FORMAT, NUMERIC_TYPE, 3),
					new Field(47, "SSIDRSV1", FIXED_FORMAT, ALPHANUMERIC_TYPE,
							5),
					new Field(52, "SSIDUSER", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 8),
					new Field(60, "SSIDCR", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1) });

	/**
	 * Start File command format.
	 */
	private static final CommandFormat SFID_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "SFIDCMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "SFIDDSN", VARIABLE_FORMAT, ALPHANUMERIC_TYPE,
							26),
					new Field(27, "SFIDRSV1", FIXED_FORMAT, ALPHANUMERIC_TYPE,
							9),
					new Field(36, "SFIDDATE", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 6),
					new Field(42, "SFIDTIME", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 6),
					new Field(48, "SFIDUSER", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 8),
					new Field(56, "SFIDDEST", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 25),
					new Field(81, "SFIDORIG", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 25),
					new Field(106, "SFIDFMT", FIXED_FORMAT, ALPHANUMERIC_TYPE,
							1),
					new Field(107, "SFIDLRECL", VARIABLE_FORMAT, NUMERIC_TYPE,
							5),
					new Field(112, "SFIDFSIZ", VARIABLE_FORMAT, NUMERIC_TYPE, 7),
					new Field(119, "SFIDREST", VARIABLE_FORMAT, NUMERIC_TYPE, 9) });

	/**
	 * Start File Positive Answer command format.
	 */
	private static final CommandFormat SFPA_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "SFPACMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "SFPAACNT", VARIABLE_FORMAT, NUMERIC_TYPE, 9) });

	/**
	 * Start File Negative Answer command format.
	 */
	private static final CommandFormat SFNA_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "SFNACMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "SFNAREAS", FIXED_FORMAT, NUMERIC_TYPE, 2),
					new Field(3, "SFNARRTR", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1) });

	/**
	 * Set Credit command format.
	 */
	private static final CommandFormat CDT_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "CDTCMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "CDTRSV1", FIXED_FORMAT, ALPHANUMERIC_TYPE, 2) });

	/**
	 * End File command format.
	 */
	private static final CommandFormat EFID_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "EFIDCMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "EFIDRCNT", VARIABLE_FORMAT, NUMERIC_TYPE, 9),
					new Field(10, "EFIDUCNT", VARIABLE_FORMAT, NUMERIC_TYPE, 12) });

	/**
	 * End File Positive Answer command format.
	 */
	private static final CommandFormat EFPA_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "EFPACMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "EFPACD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1) });

	/**
	 * End File Negative Answer command format.
	 */
	private static final CommandFormat EFNA_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "EFNACMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "EFNAREAS", FIXED_FORMAT, NUMERIC_TYPE, 2) });

	/**
	 * End Session command format.
	 */
	private static final CommandFormat ESID_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "ESIDCMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "ESIDREAS", FIXED_FORMAT, NUMERIC_TYPE, 2),
					new Field(3, "ESIDCR", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1) });

	/**
	 * Change Direction command format.
	 */
	private static final CommandFormat CD_FORMAT = new CommandFormat(
			new Field[] { new Field(0, "CDCMD", FIXED_FORMAT,
					ALPHANUMERIC_TYPE, 1) });

	/**
	 * End-to-End Response command format.
	 */
	private static final CommandFormat EERP_FORMAT = new CommandFormat(
			new Field[] {
					new Field(0, "EERPCMD", FIXED_FORMAT, ALPHANUMERIC_TYPE, 1),
					new Field(1, "EERPDSN", VARIABLE_FORMAT, ALPHANUMERIC_TYPE,
							26),
					new Field(27, "EERPRSV1", FIXED_FORMAT, ALPHANUMERIC_TYPE,
							9),
					new Field(36, "EERPDATE", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 6),
					new Field(42, "EERPTIME", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 6),
					new Field(48, "EERPUSER", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 8),
					new Field(56, "EERPDEST", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 25),
					new Field(81, "EERPORIG", VARIABLE_FORMAT,
							ALPHANUMERIC_TYPE, 25) });

	/**
	 * Ready To Receive command format.
	 */
	private static final CommandFormat RTR_FORMAT = new CommandFormat(
			new Field[] { new Field(0, "RTRCMD", FIXED_FORMAT,
					ALPHANUMERIC_TYPE, 1) });

	/**
	 * Table associating all commands format and its identifier.
	 */
	private static final Map FORMAT_MAP = new WeakHashMap(13);

	static {
		FORMAT_MAP.put(CommandIdentifier.SSRM, SSRM_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.SSID, SSID_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.SFID, SFID_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.SFPA, SFPA_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.SFNA, SFNA_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.CDT, CDT_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.EFID, EFID_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.EFPA, EFPA_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.EFNA, EFNA_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.ESID, ESID_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.CD, CD_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.EERP, EERP_FORMAT);
		FORMAT_MAP.put(CommandIdentifier.RTR, RTR_FORMAT);
	}

	/**
	 * Command Format is compound of one or more fields. This class represents
	 * the Field attributes definition.
	 */
	public static class Field {

		private final int position;

		private final String name;

		private final char format;

		private final char type;

		private final int size;

		/**
		 * This constructor create a command field with the given values, and
		 * it's intended to be used only within the <code>CommandFormat</code>
		 * class.
		 * 
		 * @param pos
		 *            Starting position of the field.
		 * @param name
		 *            Command field name.
		 * @param format
		 *            FIXED_FORMAT or VARIABLE_FORMAT.
		 * @param type
		 *            NUMERIC_TYPE or ALPHANUMERIC_TYPE.
		 * @param size
		 *            Field length.
		 */
		private Field(int pos, String name, char format, char type, int size) {
			this.position = pos;
			this.name = name;
			this.format = format;
			this.type = type;
			this.size = size;
		}

		public int getPosision() {
			return position;
		}

		public String getName() {
			return name;
		}

		public char getFormat() {
			return format;
		}

		public char getType() {
			return type;
		}

		public int getSize() {
			return size;
		}
	}

	/**
	 * Serial Version UID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Contain fields which compose the Command Format.
	 */
	private final Map fields;

	/**
	 * Count of total length of the command fields.
	 */
	private int totalLength = -1;

	/**
	 * This constructor use an array of <code>Field</code> object to determine
	 * the command format. Only for internal use.
	 */
	private CommandFormat(Field[] fieldsArray) {
		fields = buildUnmodifiableMap(fieldsArray);
	}

	/**
	 * Given an array of <code>Field</code> objects, return an unmodifiable
	 * <code>Map</code> containing the those objects as <i>Value</i> and
	 * their name property as <i>Key</i>.
	 * 
	 * @param fieldsArray
	 *            Expected array of objects.
	 * @return Unmodifiable <code>Map</code> with given objects.
	 */
	private static Map buildUnmodifiableMap(Field[] fieldsArray) {

		LinkedHashMap fieldsTable = new LinkedHashMap(fieldsArray.length);

		for (int i = 0; i < fieldsArray.length; i++) {
			Field fa = fieldsArray[i];
			fieldsTable.put(fa.getName(), fa);
		}

		return Collections.unmodifiableMap(fieldsTable);
	}

	/**
	 * Return the Command Format definition given a specific Command identifier
	 * code.
	 * 
	 * @param identifier
	 *            One of the <code>CommandIdentifier</code> static attribute.
	 * @return The corresponding Command Format definition.
	 */
	public static CommandFormat getFormat(CommandIdentifier identifier) {
		return (CommandFormat) FORMAT_MAP.get(identifier);
	}

	/**
	 * Check whether the Command Format contains a specific field by its name.
	 * 
	 * @param name
	 *            Field name.
	 * @return <code>boolean</code> indicating whether Command Format contains
	 *         this field.
	 */
	public boolean containsField(String name) {
		return fields.containsKey(name);
	}

	/**
	 * Return a Command Format field definition by its key name.
	 * 
	 * @param name
	 *            Field key name within Command Format.
	 * @return Field definition, or <code>null</code> if it does not exist.
	 */
	public Field getField(String name) {
		return (Field) fields.get(name);
	}

	/**
	 * Return the collection of fields from this Command Format.
	 * 
	 * @return <code>Collection</code>
	 */
	public Collection getFields() {
		return fields.values();
	}

	/**
	 * Get the total length of the command format, counting all fields size.
	 * 
	 * @return <code>int</code> command length.
	 */
	public int getLength() {
		if (totalLength <= 0) {
			totalLength = 0;

			Iterator it = getFields().iterator();

			while (it.hasNext()) {
				Field field = (Field) it.next();
				totalLength += field.getSize();
			}
		}

		return totalLength;
	}
}
