/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: EndSessionReason.java,v 1.5 2005/09/12 19:17:27 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.service;

import java.util.Collection;
import java.util.Map;
import java.util.WeakHashMap;

/**
 * Enumeration representing each End Session Reason avaiable.
 * <p>
 * Code table:
 * 
 * <pre>
 *          Value: 00 - Normal session termination.
 *                 01 - Command not recognised.
 *                 02 - Protocol violation.
 *                 03 - User code not known.
 *                 04 - Invalid password.
 *                 05 - Local site emergency close down.
 *                 06 - Command contained invalid data.
 *                 07 - Exchange Buffer size error.
 *                 08 - Resources not available.
 *                 09 - Time out.
 *                 10 - Mode or capabilities incompatible.
 *                 99 - Unspecified Abort code.
 * </pre>
 * 
 * @author Rafael Marins
 */
public final class EndSessionReason {

	/**
	 * Normal session termination.
	 */
	public static final EndSessionReason NORMAL_TERMINATION = new EndSessionReason(
			"00");

	/**
	 * Command not recognised.
	 * <p>
	 * An Exchange Buffer contains an invalid command identifier (1st octet of
	 * the buffer).
	 */
	public static final EndSessionReason COMMAND_NOT_RECOGNISED = new EndSessionReason(
			"01");

	/**
	 * Protocol violation.
	 * <p>
	 * An Exchange Buffer contains an invalid command for the current state of
	 * the receiver.
	 */
	public static final EndSessionReason PROTOCOL_VIOLATION = new EndSessionReason(
			"02");

	/**
	 * User code not known.
	 * <p>
	 * A Start Session (SSID) command contains an unknown or invalid
	 * Identification Code.
	 */
	public static final EndSessionReason UNKNOWN_USER_CODE = new EndSessionReason(
			"03");

	/**
	 * Invalid password.
	 * <p>
	 * A Start Session (SSID) command contains an invalid password for the
	 * specified user identification.
	 */
	public static final EndSessionReason INVALID_PASSWORD = new EndSessionReason(
			"04");

	/**
	 * Local site emergency close down.
	 * <p>
	 * The local site has entered an emergency close down mode. Communications
	 * are being forcibly terminated.
	 */
	public static final EndSessionReason EMERGENCY_CLOSE_DOWN = new EndSessionReason(
			"05");

	/**
	 * Command contained invalid data.
	 * <p>
	 * A field within a Command Exchange buffer contains invalid data.
	 */
	public static final EndSessionReason INVALID_COMMAND_DATA = new EndSessionReason(
			"06");

	/**
	 * Exchange Buffer size error.
	 * <p>
	 * The length of the Exchange Buffer as determined by the Stream
	 * Transmission Header is different to the length implied by the Command
	 * Code.
	 */
	public static final EndSessionReason EXCHANGE_BUFFER_SIZE_ERROR = new EndSessionReason(
			"07");

	/**
	 * Resources not avaiable.
	 * <p>
	 * The request for connection has been denied due to a resource shortage.
	 * The connection attempt should be retried later.
	 */
	public static final EndSessionReason RESOURCES_NOT_AVAIABLE = new EndSessionReason(
			"08");

	/**
	 * Time out.
	 */
	public static final EndSessionReason TIME_OUT = new EndSessionReason("09");

	/**
	 * Mode or capabilities are incompatible.
	 */
	public static final EndSessionReason INCOMPATIBLE_MODE = new EndSessionReason(
			"10");

	/**
	 * Unspecified Abort code.
	 * <p>
	 * An error was detected for which no specific code is defined.
	 */
	public static final EndSessionReason UNSPECIFIED_ABORT = new EndSessionReason(
			"99");

	/**
	 * Map used to store typed enums.
	 */
	private static final Map ENUM_MAP = new WeakHashMap(12);

	/*
	 * Add each typed enum to Map on first class initialization.
	 */
	static {
		addEnum(NORMAL_TERMINATION);
		addEnum(COMMAND_NOT_RECOGNISED);
		addEnum(PROTOCOL_VIOLATION);
		addEnum(UNKNOWN_USER_CODE);
		addEnum(INVALID_PASSWORD);
		addEnum(EMERGENCY_CLOSE_DOWN);
		addEnum(INVALID_COMMAND_DATA);
		addEnum(EXCHANGE_BUFFER_SIZE_ERROR);
		addEnum(RESOURCES_NOT_AVAIABLE);
		addEnum(TIME_OUT);
		addEnum(INCOMPATIBLE_MODE);
		addEnum(UNSPECIFIED_ABORT);
	}

	private String reasonCode;

	/**
	 * Enumeration constructor where reason code is specified to satisfy
	 * ReasonCode interface.
	 * 
	 * @param aReasonCode
	 */
	private EndSessionReason(String aReasonCode) {
		super();
		this.reasonCode = aReasonCode;
	}

	/**
	 * Return the protocol representation of EndSessionReason enum.
	 * 
	 * @return <code>String</code> corresponding protocol code.
	 */
	public String getCode() {
		return this.reasonCode;
	}

	/**
	 * Convenient method for parsing the proper TransferMode instance given a
	 * identifier character.
	 * 
	 * @param code
	 *            The transfer mode character being evaluated
	 * @return CommandIdentifier Instance that correspond to the parameter
	 * @throws CommandNotRecognisedException
	 *             Command not recognised
	 */
	public static EndSessionReason parse(String code) {
		return (EndSessionReason) EnumUtil.parseEnum(ENUM_MAP, code);
	}

	/**
	 * Return a collection of the avaiable typed enumerations.
	 * 
	 * @return
	 */
	public static Collection values() {
		return ENUM_MAP.values();
	}

	/**
	 * Helper method to add a typed enum to the static Map.
	 * 
	 * @param type
	 *            Typed enum.
	 */
	private static void addEnum(EndSessionReason type) {
		String code = type.getCode();
		ENUM_MAP.put(code, type);
	}
}
