/**
 *  OdetteJ - Open Source Odette FTP implementation in Java(TM).
 *  Copyright (C) 2005 Fossil E-Commerce, http://www.fossilec.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ------------------------------------------------------------------------
 *  $Id: StreamTransmissionBuffer.java,v 1.9 2005/08/15 16:43:36 rmarins Exp $
 *  ------------------------------------------------------------------------
 */

package org.fossilec.odettej.service;

import java.io.Serializable;
import java.nio.ByteBuffer;

import org.fossilec.odettej.OdetteFTPException;

/**
 * The reason of the Stream Transmission Buffer as in the specification is to
 * encapsulate the Odette Exchange Buffer in an application level flow control.
 * 
 * @author Rafael Marins
 */
public class StreamTransmissionBuffer implements Serializable {

	/**
	 * Serial Version UID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Stream Transmission Header has a fixed length at the beginning octets to
	 * each Stream Transmission Buffer.
	 */
	public static final byte HEADER_LENGTH = 4;

	/**
	 * The minimum Stream Transmission Buffer length calculation is based on the
	 * STH_LENGHT + the smallest odette exchange buffer (like CD or RTR
	 * commands).
	 */
	private static final int MIN_STB_LENGTH = 5;

	/**
	 * The maximum Stream Transmission Buffer length. Calculation is basically
	 * STH_LENGTH + Maximum Odette Exchange Buffer length determined by the
	 * protocol version being used.
	 */
	private static final int MAX_STB_LENGTH = (OdetteExchangeBuffer.MAX_OEB_LENGTH + HEADER_LENGTH);

	/**
	 * The version header field correspond to the high bits of the first buffer
	 * octet.
	 */
	private static final byte STH_VERSION = 0x10;

	/**
	 * The flags header field correpsond to the lower bits of the first buffer
	 * octet.
	 */
	private static final byte STH_FLAGS = 0x00;

	/**
	 * Stream Transmission Header is represented by an allocated buffer with a
	 * capacity equal to its relative constant length.
	 */
	private ByteBuffer headerBuffer;

	/**
	 * Being used to hold the an Exchange Buffer value of this Stream
	 * Transmission Buffer.
	 */
	private OdetteExchangeBuffer exchangeBuffer;

	/**
	 * Keep the buffer length value specified from a field in the Stream
	 * Transmission Header, representing the length of the overall Stream
	 * Transmission Buffer (header + exchange buffer).
	 */
	private int length;

	/**
	 * Constructor for enclosing an Odette Exchange Buffer into a Stream
	 * Transmission Buffer.
	 * 
	 * @param exchangeBuffer
	 *            Odette Exchange Buffer.
	 */
	public StreamTransmissionBuffer(OdetteExchangeBuffer exchangeBuffer) {
		this.exchangeBuffer = exchangeBuffer;
		setHeaderBuffer(STH_VERSION, STH_FLAGS, exchangeBuffer.getLength());
	}

	/**
	 * Constructor used to instantiate a new Stream Transmission Buffer from a
	 * raw exchange buffer coming through the transport service.
	 * 
	 * @param rawBuffer
	 *            Buffer containing the header and exchange buffer.
	 * @throws OdetteFTPException
	 * @throws BufferCorruptedException
	 *             If there is any difference between the exchange buffer length
	 *             specified in the Stream Transmission Buffer header and the
	 *             expected command size, or exceeding the range limit.
	 */
	public StreamTransmissionBuffer(ByteBuffer rawBuffer, boolean strictFormat)
			throws OdetteFTPException {

		/*
		 * The smallest STB is 5 octets consisting of a 4 octet header followed
		 * by a 1 octet Exchange Buffer such as a Change Direction (CD) command.
		 * 
		 * The maximum Exchange Buffer length that can be negotiated is 99999
		 * octets giving a STB length of 100003.
		 */
		int bufferLength = rawBuffer.remaining();

		if (bufferLength < MIN_STB_LENGTH || bufferLength > MAX_STB_LENGTH) {
			throw new BufferCorruptedException("Invalid buffer length: "
					+ bufferLength);
		}

		/*
		 * Set the Stream Transmission Header from the raw buffer data. Than
		 * buffer is positioned just after the read header octets.
		 */
		setHeaderBuffer(rawBuffer);
		rawBuffer.position(StreamTransmissionBuffer.HEADER_LENGTH);

		if (getLength() != bufferLength) {
			throw new BufferCorruptedException("Length is different to the "
					+ "specified by the Stream Transmission Header");
		}

		/*
		 * Create the corresponding OdetteExchangeBuffer instance from the
		 * remaining buffer.
		 */
		OdetteExchangeBuffer oeb = OdetteExchangeBuffer.getInstance(rawBuffer,
				strictFormat);
		setExchangeBuffer(oeb);
	}

	/**
	 * Set up the Stream Transmission Header from the buffer specified in the
	 * parameter. The buffer parameter is not distinguished between raw exchange
	 * buffer or header buffer.
	 * 
	 * Buffer is read from it first octet, and positioned at the end of header
	 * octets after the operation.
	 * 
	 * @param buffer
	 *            A buffer that contains header octets at the beginning.
	 */
	protected void setHeaderBuffer(ByteBuffer buffer) {
		int entryLimit = buffer.limit();

		buffer.limit(buffer.position() + HEADER_LENGTH);

		this.headerBuffer = buffer.slice();
		this.length = BufferUtil.parseHeaderLengthField(getHeaderBuffer());

		buffer.limit(entryLimit);
	}

	/**
	 * Set the Stream Transmission Buffer header field with the specified
	 * values.
	 * 
	 * @param headerVersion
	 *            High order bits of the first header octet indicating the
	 *            protocol version being used by the implementation.
	 * @param headerFlags
	 *            Low order bits of the first header octet indicating the flags
	 *            of the following packet.
	 * @param contentBufferLength
	 *            The Stream Transmission Buffer length.
	 */
	private void setHeaderBuffer(byte headerVersion, byte headerFlags,
			int contentBufferLength) {

		int totalLength = contentBufferLength + HEADER_LENGTH;

		/*
		 * Create the Stream Transmission Header octets. The first octet is
		 * composed with VERSION high octet bits and FLAGS lower octet bits. The
		 * lasting three octets are filled with the overall Stream Transmission
		 * Buffer length.
		 */
		byte headerOctects[] = new byte[] {
				(byte) (headerVersion | headerFlags), // version and flags
				(byte) ((totalLength & 0xff0000) >> 16), // first length
				// octet
				(byte) ((totalLength & 0xff00) >> 8), // second length octet
				(byte) ((totalLength & 0xff)) }; // thrid length octet

		setLength(totalLength);

		headerBuffer = ByteBuffer.allocateDirect(HEADER_LENGTH);
		headerBuffer.put(headerOctects);
	}

	/**
	 * Return the Stream Transmission Header buffer positioned at the first
	 * octet.
	 * 
	 * @return ByteBuffer
	 */
	public ByteBuffer getHeaderBuffer() {
		headerBuffer.rewind();
		return headerBuffer;
	}

	/**
	 * Return the Stream Transmission Buffer length. When the instance is
	 * initialized with the raw exchange buffer, this value is extracted from
	 * the buffer header length field.
	 * 
	 * @return int Stream Transmission Buffer length.
	 */
	public int getLength() {
		return length;
	}

	protected void setLength(int length) {
		this.length = length;
	}

	/**
	 * Convenient method for returning both content buffer as an array.
	 * 
	 * @return ByteBuffer[] Stream Transmission Header and the Odette Exchange
	 *         Buffer.
	 */
	public ByteBuffer[] getBuffers() {
		return new ByteBuffer[] { getHeaderBuffer(), exchangeBuffer.getBuffer() };
	}

	/**
	 * Return a Command Exchange Buffer or Data Exchange Buffer.
	 * 
	 * @return ExchangeBuffer
	 */
	public OdetteExchangeBuffer getExchangeBuffer() {
		return exchangeBuffer;
	}

	/**
	 * Set the Stream Transmission Buffer with a given Odette Exchange Buffer.
	 * 
	 * @param exchangeBuffer
	 *            Odette Exchange Buffer.
	 */
	protected void setExchangeBuffer(OdetteExchangeBuffer exchangeBuffer) {
		this.exchangeBuffer = exchangeBuffer;
	}
}